import logging
from unittest import TestCase
from requests import RequestException
from copr_common.request import SafeRequest, RequestRetryError
from . import mock


class TestStringMethods(TestCase):

    def setUp(self):
        self.url = "http://example.com/"
        self.data = {
            "foo": "bar",
            "bar": [1, 3, 5],
        }
        self.log = logging.getLogger("testlog")

    @mock.patch("copr_common.request.post")
    def test_send_request_not_200(self, post_req):
        post_req.return_value.status_code = 501
        with self.assertRaises(RequestRetryError):
            request = SafeRequest(log=self.log)
            request._send_request(self.url, "post", self.data)
        self.assertTrue(post_req.called)

    @mock.patch("copr_common.request.post")
    def test_send_request_post_error(self, post_req):
        post_req.side_effect = RequestException()
        with self.assertRaises(RequestRetryError):
            request = SafeRequest(log=self.log)
            request._send_request(self.url, "post", self.data)
        self.assertTrue(post_req.called)



"""
Streamed reading of Popen() stdout and stderr:

    proc = PosixPipedProcess(["command"])
    for chunk, output = proc.readchunks()
        assert output in ["stdout", "stderr"]
        process(chunk)
"""

from subprocess import Popen, PIPE
from threading import Thread
from queue import Queue
import select
import time

EOF = 0
CUT = 1
LOOP = 2
TIMEOUT = 3
ERR = -1

def _pipe_consumer_thread(pipe, queue, identifier, maxlen=None, poll=5.0):
    """
    Read the PIPE in a separate thread, and enqueue the chunks of output into
    the QUEUE.  The chunks are stored as:
        [(pipename, chunk), ...,  # normal chunks (of bytes)
         (pipename, EOF|CUT|ERR), # pre-termination sequence
         None,                    # iter(sentinel) (stop iteration)
        ]
    """
    counter = 0
    fd = pipe.fileno()
    try:
        while True:
            readable, _, exc = select.select([fd], [], [fd], poll)
            if exc:
                queue.put((identifier, ERR))
                break

            if not readable:
                queue.put((identifier, LOOP))
                continue

            # We need read1() to accept partial reads.
            # https://docs.python.org/3/library/io.html#io.BufferedIOBase.read1
            chunk = pipe.read1(1024)
            if not chunk:
                queue.put((identifier, EOF))
                break

            if maxlen:
                remains = maxlen - counter
                counter += len(chunk)
                if counter > maxlen:
                    chunk = chunk[:remains]
                    queue.put((identifier, chunk))
                    queue.put((identifier, CUT))
                    break

            # send a "normal" continuous chunk to the reader
            queue.put((identifier, chunk))

    finally:
        # Always notify the parent process that we quit!
        queue.put(None)


class PosixPipedProcess:
    """
    Start command using subprocess.Popen() and allow streamed reading of its
    stdout and stderr outputs.
    """
    def __init__(self, command, timeout=None, stdout_limit=None,
                 stderr_limit=None, poll=5.0, **kwargs):
        self.command = command
        self.kwargs = kwargs
        self.stdout_limit = stdout_limit
        self.stderr_limit = stderr_limit
        self.returncode = None
        self.started = None
        self._stopreason = None
        self.killed = None
        self.timeout = timeout
        self.poll = poll

    def timeouted(self):
        """ Return True if process timeouted """
        if not self.timeout:
            return False
        return time.time() - self.started > self.timeout

    def has_cut(self):
        """ Return true if the stdout_limit or stderr_limit is reached """
        return self.stopreason == CUT

    @property
    def stopreason(self):
        """
        Get the reason for stop.
        """
        return self._stopreason


    @stopreason.setter
    def stopreason(self, reason):
        """
        Set the stop reason to EOF, CUT, TIMEOUT or ERR
        """
        if self._stopreason != EOF:
            return
        self._stopreason = reason

    def readchunks(self):
        """
        (line, type)
        (line, type)
        (None: "error string") => Error
        (None: None)           => EOF
        """
        self.returncode = None
        self.started = time.time()
        self.killed = False
        self._stopreason = EOF
        que = Queue()

        # EL7: no __enter__ support in EL7
        # pylint: disable=consider-using-with
        process = Popen(self.command, stdout=PIPE, stderr=PIPE, **self.kwargs)
        tout = Thread(target=_pipe_consumer_thread,
                      args=[process.stdout, que, "stdout",
                            self.stdout_limit, self.poll])
        terr = Thread(target=_pipe_consumer_thread,
                      args=[process.stderr, que, "stderr",
                            self.stderr_limit, self.poll])
        tout.start()
        terr.start()
        def _kill():
            if self.killed:
                return
            self.killed = True
            process.kill()

        try:
            # we need to find two sentinels for both stdout/stderr
            for _ in [1, 2]:
                for fd, chunk in iter(que.get, None):
                    if self.timeouted():
                        self.stopreason = TIMEOUT
                        _kill()

                    if not isinstance(chunk, int):
                        assert chunk
                        yield (chunk, fd)
                        continue

                    if chunk == LOOP:
                        continue

                    # one of the streams ended
                    if chunk in [CUT, ERR]:
                        self.stopreason = chunk
                        _kill()
                        continue

                    assert chunk == EOF
        finally:
            # Subprocesses and threads need to finish.
            process.wait()
            terr.join()
            tout.join()
            self.returncode = process.returncode



"""
Helper methods/classes for 'copr-backend-*' scripts.
"""

import os
import sys
import argparse
import contextlib
import logging
import daemon

import setproctitle

from copr_common.redis_helpers import get_redis_connection


@contextlib.contextmanager
def nullcontext():
    """
    contextlib.nullcontext is not available in Python 3.6, but we are still
    Python 3.6+ compatible because of EL 8
    """
    yield None


class BackgroundWorker:
    """
    copr-backend-process-* abstraction
    """

    redis_logger_id = 'unknown'
    frontend_client = None
    _redis_conn = None

    def __init__(self):
        # just setup temporary stderr logger
        self.log = logging.getLogger()
        self.log.setLevel(logging.DEBUG)
        self.log.addHandler(logging.StreamHandler())

        if os.getuid() == 0:
            self.log.error("running as UID=0 (root), probably not expected")
            sys.exit(1)

        self.opts = None
        self.args = self._get_argparser().parse_args(sys.argv[1:])

    @staticmethod
    def setproctitle(text):
        """ set the process title, beginning with the script name """
        command = " ".join(sys.argv)
        setproctitle.setproctitle("{} (command: {})".format(text, command))

    @property
    def _redis(self):
        if not self._redis_conn:
            self._redis_conn = get_redis_connection(self.opts)
        return self._redis_conn

    def redis_set_worker_flag(self, flag, value=1):
        """
        Set flag in Reids DB for corresponding worker.  NO-OP if there's no
        redis connection (when run manually).
        """
        if not self.has_wm:
            return
        self._redis.hset(self.args.worker_id, flag, value)

    def redis_get_worker_flag(self, flag):
        """
        Get flag from Redis DB entry of corresponding worker.  If there's no
        Redis connection (manual run) return None
        """
        if not self.has_wm:
            return None
        return self._redis.hget(self.args.worker_id, flag)

    @classmethod
    def _get_argparser(cls):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--worker-id",
            help=("worker ID which already exists in redis DB (used by "
                  "WorkerManager only, copr-internal option)"),
        )
        parser.add_argument(
            "--daemon",
            action='store_true',
            help="execute the task on background, as daemon process"
        )
        parser.add_argument(
            "--silent",
            action='store_true',
            help="don't print logs, even when run without --daemon",
        )
        parser.add_argument(
            "--backend-config",
            help="alternative path to /etc/copr/copr-be.conf",
        )
        cls.adjust_arg_parser(parser)
        return parser

    @classmethod
    def adjust_arg_parser(cls, parser):
        """
        The inherited worker class will need more commandline options than those
        we provide by default.  Override is required.
        """
        raise NotImplementedError()

    def handle_task(self):
        """
        Abstract method for handling the task.  This should never throw any
        exception, and we don't expect it to return any value.
        """
        raise NotImplementedError()

    @property
    def has_wm(self):
        """
        Returns True if worker manager started this process, and False
        if it is manual run.
        """
        return bool(self.worker_id)

    @property
    def worker_id(self):
        """ Return worker ID if set by worker manager, or None.  """
        return self.args.worker_id

    def _wm_started(self):
        if not self.has_wm:
            return True

        self.redis_set_worker_flag('started', 1)
        self.redis_set_worker_flag('PID', os.getpid())

        data = self._redis.hgetall(self.args.worker_id)
        if 'allocated' not in data:
            self.log.error("too slow box, manager thinks we are dead")
            self._redis.delete(self.args.worker_id)
            return False

        # There's still small race on a very slow box (TOCTOU in manager, the
        # db entry can be deleted after our check above ^^).  But we don't risk
        # anything else than concurrent run of multiple workers in such case.
        return True

    def preparations_for_manager(self):
        """
        Hook called right after creating the worker process (daemonized,
        if requested) and after successful notification to Manager that
        worker started.  For manual runs (no --worker-id specified) this
        hook isn't called at all.
        """

    def _daemonized_part(self):
        # notify WorkerManager first, to minimize race window
        self._wm_started()
        if self.has_wm:
            self.preparations_for_manager()
        try:
            self.handle_task()
        except Exception as exc:  # pylint: disable=W0703
            self.log.exception("unexpected failure %s", str(exc))
            sys.exit(1)

    def process(self):
        """ process the task """
        context = nullcontext()
        if self.args.daemon:
            context = daemon.DaemonContext(umask=0o022)

        with context:
            self._daemonized_part()



"""
Copr common code related to redis
"""

from redis import StrictRedis


def get_redis_connection(opts):
    """
    Creates redis client object using backend config

    :rtype: StrictRedis
    """
    kwargs = {}
    if hasattr(opts, "redis_db"):
        kwargs["db"] = opts.redis_db
    if hasattr(opts, "redis_host"):
        kwargs["host"] = opts.redis_host
    if hasattr(opts, "redis_port"):
        kwargs["port"] = opts.redis_port
    if hasattr(opts, "redis_password"):
        kwargs["password"] = opts.redis_password

    return StrictRedis(encoding="utf-8", decode_responses=True, **kwargs)



"""
Abstract class Dispatcher for Build/Action dispatchers.
"""

import os
import time
import json
import logging
import multiprocessing
from setproctitle import setproctitle

from copr_common.redis_helpers import get_redis_connection
from copr_common.worker_manager import WorkerManager


class Dispatcher(multiprocessing.Process):
    """
    1) Fetch tasks from frontend.
    2) Fill the WorkerManager queue.
    3) Run 'WorkerManager.run()'.
    4) Go to 1)

    See also:
    https://docs.pagure.org/copr.copr/developer_documentation/dispatchers.html
    https://docs.pagure.org/copr.copr/developer_documentation/worker_manager.html
    """

    # set to either 'action' or 'build' in sub-class
    task_type = 'task_type'

    # either ActionWorkerManager or BuildWorkerManager
    worker_manager_class = WorkerManager

    # how many background workers we let the WorkerManager start, by default
    # there's no limit
    max_workers = float("inf")

    # we keep track what build's newly appeared in the task list after fetching
    # the new set from frontend after get_frontend_tasks() call
    _previous_task_fetch_ids = set()

    def __init__(self, opts):
        super().__init__(name=self.task_type + '-dispatcher')

        self.sleeptime = 0
        self.opts = opts
        self.log = logging.getLogger()
        self.frontend_client = None
        # list of applied WorkerLimit instances
        self.limits = []

    @classmethod
    def _update_process_title(cls, msg=None):
        proc_title = "{} dispatcher".format(cls.task_type.capitalize())
        if msg:
            proc_title += " - " + msg
        setproctitle(proc_title)

    def get_frontend_tasks(self):
        """
        Get _unfiltered_ list of tasks (QueueTask objects) from frontend (the
        set needs to contain both running and pending jobs).
        """
        raise NotImplementedError

    def get_cancel_requests_ids(self):
        """
        Return list of QueueTask IDS that should be canceled.
        """
        _subclass_can_use = (self)
        return []

    def report_canceled_task_id(self, task_id, was_running):
        """
        Report back to Frontend that this task was canceled.  By default this
        isn't called, so it is NO-OP by default.
        """

    def _print_added_jobs(self, tasks):
        job_ids = {task.id for task in tasks}
        new_job_ids = job_ids - self._previous_task_fetch_ids
        if new_job_ids:
            self.log.info("Got new '%s' tasks: %s", self.task_type, new_job_ids)
        self._previous_task_fetch_ids = job_ids

    def run(self):
        """
        Starts the infinite task dispatching process.
        """
        self.log.info("%s dispatching started", self.task_type.capitalize())
        self._update_process_title()

        redis = get_redis_connection(self.opts)
        worker_manager = self.worker_manager_class(
            redis_connection=redis,
            log=self.log,
            max_workers=self.max_workers,
            frontend_client=self.frontend_client,
            limits=self.limits,
        )

        timeout = self.sleeptime
        while True:
            self._update_process_title("getting tasks from frontend")
            self.log.info("getting %ss from frontend", self.task_type)
            start = time.time()

            tasks = self.get_frontend_tasks()
            if tasks:
                worker_manager.clean_tasks()

            self._print_added_jobs(tasks)
            for task in tasks:
                worker_manager.add_task(task)

            self._update_process_title("getting cancel requests")
            for task_id in self.get_cancel_requests_ids():
                was_running = worker_manager.cancel_task_id(task_id)
                self.report_canceled_task_id(task_id, was_running)

            self.dump_queue(worker_manager)

            # process the tasks
            self._update_process_title("processing tasks")
            worker_manager.run(timeout=timeout)

            sleep_more = timeout - (time.time() - start)
            if sleep_more > 0:
                time.sleep(sleep_more)

        # reset the title
        self._update_process_title()

    def dump_queue(self, worker_manager):
        """
        Dump the current priority queue state into a JSON file.
        This is useful when we want to see the exact state of the queue from
        backend's perspective (e.g. when some tasks are not being processed).
        """
        # This function is usefull for debugging, there is no need to run it
        # all the time
        if self.log.level > logging.DEBUG:
            return

        queue = []
        for entry in worker_manager.tasks.prio_queue:
            # This should never happen but let's make sure we don't traceback
            if not isinstance(entry, list) or len(entry) != 3:
                return

            priority, count, task = entry
            queue.append([priority, count, task.id])

        name = "copr-{0}-queue.json".format(self.name)
        path = os.path.join(self.opts.log_dir, name)
        with open(path, "w", encoding="utf8") as fp:
            json.dump(queue, fp, indent=4)
        self.log.debug("Priority queue dumped to: %s", path)



"""
Helper methods that can be used by any part of the Copr stack.  Typically small
things that don't need to have it's own file.
"""

import os
import pwd
import sys


# When build is resubmitted with user SSH,
# how long the builder is kept alive if user doesn't prolong it (in seconds).
USER_SSH_DEFAULT_EXPIRATION = 60 * 60

# When build is resubmitted with user SSH,
# how long the builder can be prolonged to be kept alive (in seconds).
USER_SSH_MAX_EXPIRATION = 60 * 60 * 24 * 2

# When build is resubmitted with user SSH,
# in what file the expiration timestamp should be stored
USER_SSH_EXPIRATION_PATH = "/run/copr-builder-expiration"


def script_requires_user(username):
    """
    Exit if the current system user name doesn't equal to the USERNAME argument.
    """
    actual_username = pwd.getpwuid(os.getuid())[0]
    if actual_username != username:
        msg = (
            "This script should be executed "
            "as '{0}' user, not '{1}'\n"
        ).format(
            username,
            actual_username,
        )
        sys.stderr.write(msg)
        sys.exit(1)


def chroot_to_branch(chroot):
    """
    Get a git branch name from chroot. Follow the fedora naming standard.
    """
    name, version, _arch = chroot.rsplit("-", 2)
    if name == "fedora":
        if version == "rawhide":
            return "master"
        abbrev = "f"
    elif name == "epel" and int(version) <= 6:
        abbrev = "el"
    elif name == "mageia" and version == "cauldron":
        abbrev = "cauldron"
        version = ""
    elif name == "mageia":
        abbrev = "mga"
    return "{}{}".format(abbrev, version)


def timedelta_to_dhms(delta):
    """
    By default the `datetime.timedelta` provides only days and seconds. Minutes,
    hours, and the human friendly number of seconds, needs to be calculated.
    """
    days, remainder = divmod(delta.total_seconds(), 24 * 60 * 60)
    hours, remainder = divmod(remainder, 60 * 60)
    minutes, seconds = divmod(remainder, 60)
    return int(days), int(hours), int(minutes), int(seconds)



import sys
import pytest

if sys.version_info[0] >= 3:

    from copr_common.subprocess_live_output import PosixPipedProcess

    @pytest.mark.parametrize('output', ['stdout', 'stderr'])
    def test_posix_live_output_one_stream(output):
        redirect = "" if output == 'stdout' else '>&2'
        opposite = "stderr" if output == 'stdout' else 'stdout'
        proc = PosixPipedProcess([
            "sh", "-c",
            "echo {redirect} {output} ; sleep 0.2; "
            "echo {redirect} -n {output}".format(output=output, redirect=redirect)
        ])
        outs = {"stdout": "", "stderr": ""}
        for chunk, out in proc.readchunks():
            assert out != opposite
            outs[out] += chunk.decode("utf8")

        assert proc.returncode == 0
        assert proc.stopreason == 0
        assert outs[opposite] == ""
        assert outs[output] == "{output}\n{output}".format(output=output)

    def test_posix_live_output_both():
        proc = PosixPipedProcess([
            "sh", "-c",
            "echo -n output ; "
            "echo -n errout >&2 ; "
            "echo output ; "
            "echo errout >&2"
        ])
        outs = {"stdout": "", "stderr": ""}
        for chunk, out in proc.readchunks():
            outs[out] += chunk.decode("utf8")

        assert proc.returncode == 0
        assert proc.stopreason == 0
        assert not proc.has_cut()
        assert outs['stdout'] == "outputoutput\n"
        assert outs['stderr'] == "errouterrout\n"


    @pytest.mark.parametrize("dataset", [
        (["/bin/false"], 1, "", ""),
        (["/bin/true"], 0, "", ""),
    ])
    def test_posix_live_output_exit_status(dataset):
        cmd, exp_rc, exp_out, exp_err = dataset
        proc = PosixPipedProcess(cmd)
        outs = {"stdout": "", "stderr": ""}
        for chunk, out in proc.readchunks():
            outs[out] += chunk.decode("utf8")
        assert proc.returncode == exp_rc
        assert proc.stopreason == 0
        assert not proc.has_cut()
        assert outs['stdout'] == exp_out
        assert outs['stderr'] == exp_err


    def test_posix_live_output_timeout():
        cmd = ["bash", "-c", "echo ahoj; sleep 100"]
        proc = PosixPipedProcess(cmd, timeout=2, poll=0.1)
        outs = {"stdout": "", "stderr": ""}
        for chunk, out in proc.readchunks():
            outs[out] += chunk.decode("utf8")
        assert proc.timeouted()
        assert not proc.has_cut()
        assert proc.returncode == -9
        assert outs['stdout'] == "ahoj\n"
        assert outs['stderr'] == ""


    def test_posix_live_output_cut():
        cmd = ["bash", "-c", "echo aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"]
        proc = PosixPipedProcess(cmd, stdout_limit=10)
        outs = {"stdout": "", "stderr": ""}
        for chunk, out in proc.readchunks():
            outs[out] += chunk.decode("utf8")
        assert proc.returncode == 0
        assert proc.has_cut()
        assert outs['stdout'] == "aaaaaaaaaa"
        assert outs['stderr'] == ""

    def test_posix_live_output_cut_long():
        cmd = ["bash", "-c", "while :; do echo -n aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa >&2; done"]
        proc = PosixPipedProcess(cmd, stderr_limit=10)
        outs = {"stdout": "", "stderr": ""}
        for chunk, out in proc.readchunks():
            outs[out] += chunk.decode("utf8")
        assert outs['stdout'] == ""
        assert outs['stderr'] == "aaaaaaaaaa"
        assert proc.returncode == -9
        assert proc.has_cut()



"""
WorkerManager for processing a priority queue and starting background jobs
"""

import os
import time
from heapq import heappop, heappush
import itertools
import logging
import subprocess


class WorkerLimit:
    """
    Limit for the number of tasks being processed concurrently

    WorkerManager expects that it's caller fills the task queue only with tasks
    that should be processed.  Then WorkerManager is completely responsible for
    sorting out the queue, and behave -> respect the given limits.

    WorkerManager implements really stupid algorithm to respect the limits;  we
    simply drop the task from queue (and continue to the next task) if it was
    about to exceed any given limit.  This is the only option at this moment
    because JobQueue doesn't allow us to skip some task, and return to it later.
    It is not a problem for Copr dispatchers though because we re-add the
    dropped tasks to JobQueue anyways -- after the next call to the
    Dispatcher.get_frontend_tasks() method (see "sleeptime" configuration
    option).

    Each Limit object works as a statistic counter for the list of _currently
    processed_ tasks (i.e. not queued tasks!).  And we may want to query the
    statistics anytime we want to.  Both calculating and querying the statistics
    must be as fast as possible, therefore the interface only re-calculates the
    stats when WorkerManager starts/stops working on some task.

    One may wonder why to bother with limits, and why not to delegate this
    responsibility on WorkerManager caller (IOW don't put the task to queue if
    it is not yet the right time process it..).  That would be ideal case, but
    at the time of filling the queue caller has no idea about the currently
    running BackgroundWorker instances (those need to be calculated to
    statistics, too).
    """

    def __init__(self, name=None):
        self._name = name

    def worker_added(self, worker_id, task):
        """ Add worker and it's task to statistics.  """
        raise NotImplementedError

    def check(self, task):
        """ Check if the task can be added without crossing the limit. """
        raise NotImplementedError

    def clear(self):
        """ Clear the statistics. """
        raise NotImplementedError

    def info(self):
        """ Get the user-readable info about the limit object """
        if self._name:
            return "'{}'".format(self._name)
        return "Unnamed '{}' limit".format(type(self).__name__)


class PredicateWorkerLimit(WorkerLimit):
    """
    Calculate how many tasks being processed by currently running workers match
    the given predicate.
    """
    def __init__(self, predicate, limit, name=None):
        """
        :param predicate: function object taking one QueueTask argument, and
            returning True or False
        :param limit: how many tasks matching the ``predicate`` are allowed to
            be processed concurrently.
        """
        super().__init__(name)
        self._limit = limit
        self._predicate = predicate
        self.clear()

    def clear(self):
        self._refs = {}

    def worker_added(self, worker_id, task):
        if not self._predicate(task):
            return
        self._refs[worker_id] = True

    def check(self, task):
        if not self._predicate(task):
            return True
        return len(self._refs) < self._limit

    def info(self):
        text = super().info()
        matching = ', '.join(self._refs.keys())
        if not matching:
            return text
        return "{}, matching: {}".format(text, matching)


class StringCounter:
    """
    Counter for string occurrences.  When string is None, we don't count it
    """
    def __init__(self):
        self._counter = {}

    def add(self, string):
        """ Add string to counter """
        if string is None:
            return
        if string in self._counter:
            self._counter[string] += 1
        else:
            self._counter[string] = 1

    def count(self, string):
        """ Return number ``string`` occurrences """
        return self._counter.get(string, 0)

    def __str__(self):
        items = ["{}={}".format(key, value)
                 for key, value in self._counter.items()]
        return ", ".join(items)


class HashWorkerLimit(WorkerLimit):
    """
    Assign tasks to groups per the return value of the HASHER(TASK) method.  Set
    maximum number of workers **per each such group**.
    """
    def __init__(self, hasher, limit, name=None):
        """
        :param hasher: function object taking one QueueTask argument, and
            returning string key (name of the ``group``).
        :param limit: how many tasks in the ``group`` are allowed to be
            processed at the same time.
        """
        super().__init__(name)
        self._limit = limit
        self._hasher = hasher
        self.clear()

    def clear(self):
        self._groups = StringCounter()
        self._refs = {}

    def worker_added(self, worker_id, task):
        # remember it
        group_name = self._refs[worker_id] = self._hasher(task)
        # count it
        self._groups.add(group_name)

    def check(self, task):
        group_name = self._hasher(task)
        return self._groups.count(group_name) < self._limit

    def info(self):
        text = super().info()
        return "{}, counter: {}".format(text, str(self._groups))


class JobQueue():
    """
    Priority "task" queue for WorkerManager.  Taken from:
    https://docs.python.org/3/library/heapq.html#priority-queue-implementation-notes
    The higher the 'priority' is, the later the task is taken.
    """

    def __init__(self, removed='<removed-task>'):
        self.prio_queue = []             # list of entries arranged in a heap
        self.entry_finder = {}           # mapping of tasks to entries
        self.removed = removed           # placeholder for a removed task
        self.counter = itertools.count() # unique sequence count

    def add_task(self, task, priority=0):
        'Add a new task or update the priority of an existing task'
        if repr(task) in self.entry_finder:
            self.remove_task(task)
        count = next(self.counter)
        entry = [priority, count, task]
        self.entry_finder[repr(task)] = entry
        heappush(self.prio_queue, entry)

    def remove_task(self, task):
        'Mark an existing task as removed.  Raise KeyError if not found.'
        self.remove_task_by_id(repr(task))

    def remove_task_by_id(self, task_id):
        """
        Using task id, drop the task from queue.  Raise KeyError if not found.
        """
        entry = self.entry_finder.pop(task_id)
        entry[-1] = self.removed

    def pop_task(self):
        'Remove and return the lowest priority task. Raise KeyError if empty.'
        while self.prio_queue:
            _priority, _count, task = heappop(self.prio_queue)
            if task is not self.removed:
                del self.entry_finder[repr(task)]
                return task
        raise KeyError('pop from an empty priority queue')


class QueueTask:
    """
    A base class for tasks processed by `Dispatcher` implementations
    """

    def __repr__(self):
        return str(self.id)

    @property
    def id(self):
        """
        Unique ID for distinguishing tasks
        """
        raise NotImplementedError

    @property
    def priority(self):
        """
        The higher the 'priority' is, the later the task is taken.
        """
        return 0


class WorkerManager():
    """
    Automatically process 'self.tasks' priority queue, and start background jobs
    to handle them.

    :cvar worker_prefix: Unique string across all the WorkerManager child
            classes, this is used as prefix for the workers in redis database
            and to easily determine to which WorkerManager the particlar worker
            belongs to.  So it can be anything reasonable, just make sure it is
            unique.
    :cvar worker_timeout_start: The time period we give the background process
            successfully start and identify itself (see has_worker_started()
            method).  If the background process isn't indentified after this
            timeout, we drop it from database and consider it failed.  And the
            task is re-scheduled.  Float value in seconds.
    :cvar worker_timeout_deadcheck: If the worker successfully identified itself
            after start (the has_worker_started() returns True) we know that the
            worker process at least started.  But after worker_timeout_deadcheck
            timeout we'll also keep an eye on the process by asking the
            is_worker_alive() method - whether the task is really still doing
            something on background or not (== unexpected failure cleanup).
            Fill float value in seconds.
    :cvar worker_cleanup_period: How often should WorkerManager try to cleanup
            workers? (value is a period in seconds)
    """

    # pylint: disable=too-many-instance-attributes

    worker_prefix = 'worker' # make sure this is unique in each class
    worker_timeout_start = 30
    worker_timeout_deadcheck = 3*60
    worker_cleanup_period = 3.0


    def __init__(self, redis_connection=None, max_workers=8, log=None,
                 frontend_client=None, limits=None):
        self.tasks = JobQueue()
        self.log = log if log else logging.getLogger()
        self.redis = redis_connection
        self.max_workers = max_workers
        self.frontend_client = frontend_client
        # We have to frequently ask for the actually tracked list of workers —
        # therefore we keep it here to not re-query the list from Redis all the
        # time.  We have to load the list from Redis initially, when the process
        # starts (Manager/Dispatcher class is loaded) because we want the logic
        # to survive server restarts (we adopt the old background workers).
        self._tracked_workers = set(self.worker_ids())
        self._limits = limits or []
        self._last_worker_cleanup = None

    def start_task(self, worker_id, task):
        """
        Start background job using the 'task' object taken from the 'tasks'
        queue.  The background task should _on its own_ and ASAP let the manager
        know that it successfully started (e.g. mark the job 'started' in redis
        DB), so the has_worker_started() method later gives us valid info.
        That's btw. the reason why we have the mandatory `worker_id` argument
        here, the background worker needs to know what key to update in redis.
        """
        raise NotImplementedError

    def finish_task(self, worker_id, task_info):
        """
        This is called once the worker manager consider the task to be done,
        because the `has_worker_ended()` method already returns True.  Override
        this function and use it to let Frontend know that the task finished,
        and how (whether it succeeded, etc.).
        """
        raise NotImplementedError

    def has_worker_started(self, worker_id, task_info):
        """
        The background task process should somehow notify manager that it
        already started (so we can have has_worker_started() implemented).
        By default we expect it sets 'started' attribute in redis DB, but feel
        free to override this method and invent different notification
        mechanism.
        """
        # pylint: disable=no-self-use
        # pylint: disable=unused-argument
        return 'started' in task_info

    def has_worker_ended(self, worker_id, task_info):
        """
        Check 'task_info' (dictionary output from redis) whether the task is
        already finished by worker.  If yes, do whatever is needed with the
        result (contact frontend) and return True.  If the task is still
        processed, return False.  By default we just check for the ``status``
        presence in ``task_info``, but this method is to be overridden.
        """
        _subclass_can_use = (self, worker_id)
        return 'status' in task_info

    def is_worker_alive(self, worker_id, task_info):
        """
        Check staled jobs on background, whether they haven't died before they
        notified us about the status.  We'll keep asking after
        worker_timeout_deadcheck seconds left since we tried to spawn the
        worker.

        By default we check for 'PID' presence in task_info, but users are
        advised to override this method when necessary.  We keep the 'self'
        and 'worker_id' arguments as they might be useful.
        """
        # pylint: disable=unused-argument,no-self-use
        if not 'PID' in task_info:
            return False
        pid = int(task_info['PID'])
        try:
            # Send signal=0 to the process to check whether it still exists.
            # This is just no-op if the signal was successfully delivered to
            # existing process, otherwise exception is raised.
            os.kill(pid, 0)
        except OSError:
            return False
        return True

    def get_worker_id(self, task_id):
        """
        Given the unique task representation form (usually ID), generate worker
        identificator (redis key).
        """
        return '{}:{}'.format(self.worker_prefix, task_id)

    def get_task_id_from_worker_id(self, worker_id):
        """
        Given the unique task representation form (usually ID), generate worker
        identificator (redis key).
        """
        prefix, task_id = worker_id.rsplit(':', 1)
        assert prefix == self.worker_prefix
        return task_id

    def _calculate_limits_for_task(self, worker_id, task):
        for limit in self._limits:
            limit.worker_added(worker_id, task)

    def cancel_request_done(self, task):
        """ Report back to frontend that the cancel request was finished. """

    def add_task(self, task):
        """
        Add task to queue.
        """
        task_id = repr(task)
        worker_id = self.get_worker_id(task_id)

        if worker_id in self._tracked_workers:
            # No need to re-add this to queue, but we need to calculate
            # it into the limits.
            self._calculate_limits_for_task(worker_id, task)
            self.log.debug("Task %s already has a worker process", task_id)
            return

        self.log.debug("Adding task %s to queue, priority %s", task_id,
                       task.priority)
        self.tasks.add_task(task, task.priority)

    def _drop_task_id_safe(self, task_id):
        try:
            self.tasks.remove_task_by_id(task_id)
        except KeyError:
            pass

    def cancel_task_id(self, task_id):
        """
        Using task_id, cancel corresponding task, and request worker
        shut-down (when already started)

        :return: True if worker is running on background, False otherwise
        """
        self._drop_task_id_safe(task_id)
        worker_id = self.get_worker_id(task_id)
        if worker_id not in self.worker_ids():
            self.log.info("Cancel request, worker %s is not running", worker_id)
            return False
        self.log.info("Cancel request, worker %s requested to cancel",
                      worker_id)
        self.redis.hset(worker_id, 'cancel_request', 1)
        return True

    def worker_ids(self):
        """
        Return the redis keys representing workers running on background.
        """
        return self.redis.keys(self.worker_prefix + ':*')

    def run(self, timeout=float('inf')):
        """
        Process the task (priority) queue.
        """
        now = None
        start_time = time.time()
        self.log.debug("Worker.run() start at time %s", start_time)

        # Make sure _cleanup_workers() has some effect during the run() call.
        # This is here mostly for the test-suite, because in the real use-cases
        # the worker_cleanup_period is much shorter period than the timeout and
        # the cleanup is done _several_ times during the run() call.
        self._last_worker_cleanup = 0.0

        while True:
            now = start_time if now is None else time.time()

            if not now - start_time < timeout:
                break

            self._cleanup_workers(now)

            worker_count = len(self._tracked_workers)
            if worker_count >= self.max_workers:
                self.log.debug("Worker count on a limit %s", worker_count)
                time.sleep(1)
                continue

            # We can allocate some workers, if there's something to do.
            try:
                task = self.tasks.pop_task()
            except KeyError:
                # Empty queue!
                if worker_count:
                    # It still makes sense to cycle to finish the workers.
                    self.log.debug("No more tasks, waiting for workers")
                    time.sleep(1)
                    continue
                # Optimization part, nobody is working now, and there's nothing
                # to do.  Just simply wait till the end of the cycle.
                break

            break_on_limit = False
            for limit in self._limits:
                # just skip this task, it will be processed in the next
                # run because we keep re-filling the queue
                if not limit.check(task):
                    self.log.debug("Task '%s' skipped, limit info: %s",
                                   task.id, limit.info())
                    break_on_limit = True
                    break
            if break_on_limit:
                continue

            self._start_worker(task, now)

        self.log.debug("Reaped %s processes", self._clean_daemon_processes())
        self.log.debug("Worker.run() stop at time %s", time.time())

    def _start_worker(self, task, time_now):
        worker_id = self.get_worker_id(repr(task))
        self.redis.hset(worker_id, 'allocated', time_now)
        self._tracked_workers.add(worker_id)
        self.log.info("Starting worker %s, task.priority=%s", worker_id,
                      task.priority)
        self._calculate_limits_for_task(worker_id, task)
        self.start_task(worker_id, task)

    def clean_tasks(self):
        """
        Remove all tasks from queue.
        """
        self.tasks = JobQueue()
        for limit in self._limits:
            limit.clear()

    def _delete_worker(self, worker_id):
        self.redis.delete(worker_id)
        self._tracked_workers.discard(worker_id)

    def _cleanup_workers(self, now):
        """
        Go through all the tracked workers and check if they already finished,
        failed to start or died in the background.
        """

        # This method is called very frequently (several hundreds per second,
        # for each of the attempts to start a worker in the self.run() method).
        # Because the likelihood that some of the background workers changed
        # state is pretty low, we control the frequency of the cleanup here.
        now = time.time()
        if now - self._last_worker_cleanup < self.worker_cleanup_period:
            return

        self.log.debug("Trying to clean old workers")
        self._last_worker_cleanup = time.time()

        for worker_id in self.worker_ids():
            info = self.redis.hgetall(worker_id)

            allocated = info.get('allocated', None)
            if not allocated:
                # In worker manager, we _always_ add 'allocated' tag when we
                # start worker.  So this may only happen when worker is
                # orphaned for some reason (we gave up with him), and it still
                # touches the database on background.
                self.log.info("Missing 'allocated' flag for worker %s", worker_id)
                self._delete_worker(worker_id)
                continue

            allocated = float(allocated)

            if self.has_worker_ended(worker_id, info):
                # finished worker
                self.log.info("Finished worker %s", worker_id)
                self.finish_task(worker_id, info)
                self._delete_worker(worker_id)
                continue

            if info.get('delete'):
                self.log.warning("worker %s deleted", worker_id)
                self._delete_worker(worker_id)
                continue

            if not self.has_worker_started(worker_id, info):
                if now - allocated > self.worker_timeout_start:
                    # This worker failed to start?
                    self.log.error("worker %s failed to start", worker_id)
                    self._delete_worker(worker_id)
                continue

            checked = info.get('checked', allocated)

            if now - float(checked) > self.worker_timeout_deadcheck:
                self.log.info("checking worker %s", worker_id)
                self.redis.hset(worker_id, 'checked', now)
                if self.is_worker_alive(worker_id, info):
                    continue
                self.log.error("dead worker %s", worker_id)

                # The worker could finish in the meantime, make sure we
                # hgetall() once more.
                self.redis.hset(worker_id, 'delete', 1)

    def start_daemon_on_background(self, command, env=None):
        """
        The daemon.DaemonContext() is pretty slow thing, it may take up to 1s
        to finish and return the exit status to the parent process.  But if the
        calling logic is properly prepared for potential startup failures, we
        don't have to wait at all and we can start the background process on
        background, too.  Typical work-around for starting the
        'copr-backend-process-*' scripts that are based on the
        BackgroundWorker.process() logic.
        """
        # pylint: disable=consider-using-with
        process = subprocess.Popen(command, env=env)
        self.log.debug("background pid=%s started (%s)", process.pid, command)

    def _clean_daemon_processes(self):
        """
        Wait for all the finished subprocesses to avoid the leftover zombies.
        Return the number of successfully waited processes.  Complements the
        start_daemon_on_background() above, but called automatically.
        """
        counter = 0
        try:
            # Wait for any background process (pid=-1), and no additional
            # options are needed (options=0).  All the background processes
            # should quit relatively fast (within one second).
            while True:
                (pid, _) = os.waitpid(-1, 0)
                self.log.debug("Worker Manager waited for pid=%s", pid)
                counter += 1
        except ChildProcessError:
            pass
        return counter
